import pandas as pd

from sttn.network import SpatioTemporalNetwork
from sttn.utils import get_edges_with_centroids
from keplergl import KeplerGl
from typing import Optional, List, Tuple

from .keplergl.map import MapConfig

DEFAULT_MAP_HEIGHT = 600


def choropleth(data: SpatioTemporalNetwork, node_layers: Optional[List[str]] = None, include_edges: bool = False,
               edge_size_column: Optional[str] = None) -> KeplerGl:
    map_conf = MapConfig(data_id='nodes')
    nodes = data.nodes

    node_layers = _get_node_layers(node_layers, list(nodes.columns))
    for column in node_layers:
        map_conf.add_geo_layer(label=f"node: {column}", color_column=column)
    map_data = {'nodes': nodes.copy()}

    if include_edges:
        edges = get_edges_with_centroids(data)
        map_conf.with_data_id('edges').add_arc_layer(label=f"edge: {edge_size_column}", origin_lat="lat_from",
                                                     origin_lng="long_from",
                                                     destination_lat="lat_to", destination_lng="long_to",
                                                     size_column=edge_size_column)
        map_data['edges'] = edges

    output_map = KeplerGl(height=DEFAULT_MAP_HEIGHT, data=map_data, config=map_conf.to_dict())
    return output_map


def transaction_heatmap(data: SpatioTemporalNetwork, weight_column: str, transaction_node: str = "origin",
                        time_motion_column: Optional[str] = None) -> KeplerGl:
    map_conf = MapConfig(data_id='edges')
    edges = get_edges_with_centroids(data)

    if transaction_node == "origin":
        lat = "lat_from"
        lng = "long_from"
    elif transaction_node == "destination":
        lat = "lat_to"
        lng = "long_to"
    else:
        raise ValueError(f"{transaction_node} is not one of supported values [origin, destination]")

    map_conf.add_heatmap_layer(label=f"Edge {transaction_node} {weight_column}", lat_column=lat, lng_column=lng,
                               weight_column=weight_column)
    if time_motion_column is not None:
        window = _get_time_range(data.edges[time_motion_column])
        map_conf.add_time_range_filter(time_column=time_motion_column, window=window, y_axis_column=weight_column)

    map_data = {'edges': edges}
    output_map = KeplerGl(height=DEFAULT_MAP_HEIGHT, data=map_data, config=map_conf.to_dict())
    return output_map


def transaction(data: SpatioTemporalNetwork, edge_size_column: str,
                time_motion_column: Optional[str] = None) -> KeplerGl:
    map_conf = MapConfig(data_id='edges')
    edges = get_edges_with_centroids(data)
    map_conf.add_arc_layer(label=f"edge: {edge_size_column}", origin_lat="lat_from",
                           origin_lng="long_from",
                           destination_lat="lat_to", destination_lng="long_to",
                           size_column=edge_size_column)

    if time_motion_column is not None:
        window = _get_time_range(data.edges[time_motion_column])
        map_conf.add_time_range_filter(time_column=time_motion_column, window=window, y_axis_column=edge_size_column)

    map_data = {'edges': edges}
    output_map = KeplerGl(height=DEFAULT_MAP_HEIGHT, data=map_data, config=map_conf.to_dict())
    return output_map


def _get_time_range(time_column: pd.Series) -> Tuple[int, int]:
    window_start = time_column.min()
    window_end = time_column.max()
    window = (window_start.value // 1000000, window_end.value // 1000000)
    return window


def _get_node_layers(node_layers: Optional[List[str]], node_columns: List[str]) -> List[str]:
    if node_layers is None:
        node_columns.remove('geometry')
        return node_columns
    else:
        for layer in node_layers:
            if layer not in node_columns:
                raise ValueError(f"{layer} is not one of the data node data layers: {node_layers}")

        return node_layers



import pandas as pd
import geopandas as gpd


def get_state_codes() -> pd.DataFrame:
    url = 'https://www2.census.gov/geo/docs/reference/state.txt'
    columns_to_rename = {'STATE': 'fips_code', 'STUSAB': 'usps_code', 'STATE_NAME': 'state', 'STATENS': 'gnisid'}
    data = pd.read_csv(url, delimiter='|').rename(columns=columns_to_rename)
    return data


def get_tract_geo(state: str, year: int) -> gpd.GeoDataFrame:
    state_codes = get_state_codes()
    fips_code = state_codes[state_codes.usps_code == state.upper()].iloc[0].fips_code
    url = 'https://www2.census.gov/geo/tiger/TIGER{year}/TRACT/tl_{year}_{state_fips:02d}_tract.zip'.format(
        year=year, state_fips=fips_code)
    df = gpd.read_file(url)
    df.GEOID = df.GEOID.astype(int)
    return df



from datetime import datetime
from dateutil.relativedelta import relativedelta
import pandas as pd
import geopandas as gpd
import pyarrow as pa
import pyarrow.parquet
import os

from sttn import network
from .data_provider import DataProvider

TAXI_ZONE_SHAPE_URL = 'https://data.cityofnewyork.us/api/geospatial/d3c5-ddgc?method=export&format=Shapefile'


class NycTaxiDataProvider(DataProvider):

    @staticmethod
    def build_network(taxi_trips, taxi_zones) -> network.SpatioTemporalNetwork:
        edges = taxi_trips.rename(
            columns={'PULocationID': 'origin', 'DOLocationID': 'destination', 'tpep_pickup_datetime': 'time'})
        edges_casted = edges.astype({'origin': 'int64', 'destination': 'int64'})
        taxi_zones = taxi_zones.rename(columns={'objectid': 'id'}).astype({'id': 'int32'})
        taxi_zones = taxi_zones.set_index('id')
        return network.SpatioTemporalNetwork(nodes=taxi_zones, edges=edges_casted)

    def get_data(self, taxi_type: str, month: str) -> network.SpatioTemporalNetwork:
        url = f'https://d37ci6vzurychx.cloudfront.net/trip-data/{taxi_type}_tripdata_{month}.parquet'
        taxi_data = self.cache_file(url)
        column_names = ['PULocationID', 'DOLocationID', 'tpep_pickup_datetime', 'passenger_count', 'fare_amount']
        df = pd.read_parquet(taxi_data, columns=column_names)
        df = df[(df['PULocationID'] > 0) & (df['PULocationID'] < 264)]
        df = df[(df['DOLocationID'] > 0) & (df['DOLocationID'] < 264)]
        df = df.dropna()

        from_date = datetime.strptime(month, '%Y-%m')
        to_date = from_date + relativedelta(months=1)
        df = df[(df['tpep_pickup_datetime'] >= from_date) & (df['tpep_pickup_datetime'] <= to_date)]
        df['passenger_count'] = df['passenger_count'].astype(int)
        labels = gpd.read_file(TAXI_ZONE_SHAPE_URL)
        return self.build_network(df, labels)


class Service311RequestsDataProvider(DataProvider):

    def build_network(self, requests, nyc_zip_shape):
        requests['from'] = requests['Incident Zip']
        requests['to'] = requests['Incident Zip']
        column_map = {'Latitude': 'latitude', 'Longitude': 'longitude', 'Complaint Type': 'complaint_type',
                      'Created Date': 'time', 'City': 'city'}
        requests = requests.rename(columns=column_map)
        requests = requests.drop('Incident Zip', 1)

        node_labels = nyc_zip_shape[['ZIPCODE', 'COUNTY', 'PO_NAME', 'geometry']]
        node_labels.columns = node_labels.columns.str.lower()
        node_labels = node_labels.rename(columns={'zipcode': 'id'})
        node_labels = node_labels.set_index('id')
        return network.SpatioTemporalNetwork(requests, node_labels=node_labels)

    def get_data(self, from_date, to_date):
        data = self.cache_file('https://data.cityofnewyork.us/api/views/erm2-nwe9/rows.csv')
        column_names = ['Incident Zip', 'City', 'Latitude', 'Longitude', 'Complaint Type', 'Created Date']
        filtered_file = self.filter_requests(data, from_date, to_date, column_names)

        nyc_shape = gpd.read_file(
            'https://data.cityofnewyork.us/api/views/i8iw-xf4u/files/YObIR0MbpUVA0EpQzZSq5x55FzKGM2ejSeahdvjqR20?filename=ZIP_CODE_040114.zip')
        nyc_shape['ZIPCODE'] = nyc_shape['ZIPCODE'].astype(int)
        requests = pd.read_parquet(filtered_file)
        requests['Incident Zip'] = requests['Incident Zip'].astype(int)

        return self.build_network(requests, nyc_shape)

    def filter_requests(self, requests_file, from_date, to_date, column_names):
        arg_hash = self.hash_args(from_date=from_date.timestamp(), to_date=to_date.timestamp(),
                                  column_names=column_names)
        local_filename = arg_hash + '.parquet'
        file_path = os.path.join(self.cache_dir(), local_filename)

        if not os.path.exists(file_path):
            pqwriter = None
            # write to a temporary file in order to avoid incomplete results
            tmp_file_name = arg_hash + '_tmp.parquet'
            tmp_file_path = os.path.join(self.cache_dir(), tmp_file_name)
            if os.path.exists(tmp_file_path):
                os.remove(tmp_file_path)
            types = {'Incident Zip': pd.StringDtype(), 'Complaint Type': pd.StringDtype(), 'City': pd.StringDtype()}
            chunks_iter = pd.read_csv(requests_file, parse_dates=['Created Date'], usecols=column_names, dtype=types,
                                      index_col=False, iterator=True, chunksize=1 << 18)
            for chunk in chunks_iter:
                filtered_chunk = chunk[(chunk['Created Date'] >= from_date) & (chunk['Created Date'] <= to_date)]
                filtered_chunk = filtered_chunk.dropna(subset=['Incident Zip'])
                table = pa.Table.from_pandas(df=filtered_chunk)
                if not pqwriter:
                    pqwriter = pa.parquet.ParquetWriter(tmp_file_path, table.schema)
                pqwriter.write_table(table)

            # close the parquet writer
            if pqwriter:
                pqwriter.close()
            os.rename(tmp_file_path, file_path)

        return file_path



from abc import ABC, abstractmethod
from typing import Any, Dict
import random
import string

HIGHLIGHT_COLOR = [252, 242, 26, 255]
COLOR_RANGE = {'name': 'Global Warming',
               'type': 'sequential',
               'category': 'Uber',
               'colors': ['#5A1846',
                          '#900C3F',
                          '#C70039',
                          '#E3611C',
                          '#F1920E',
                          '#FFC300']}
DEFAULT_HEATMAP_RADIUS = 10


class MapLayer(ABC):
    _id: str

    @property
    def layer_id(self):
        return self._id

    @staticmethod
    def generate_id() -> str:
        return ''.join(random.choices(string.ascii_lowercase, k=5))

    @property
    @abstractmethod
    def layer_type(self) -> str:
        return NotImplemented

    @property
    @abstractmethod
    def config(self) -> Dict[str, Any]:
        return NotImplemented

    @property
    @abstractmethod
    def visual_channels(self) -> Dict[str, Any]:
        return NotImplemented

    def get_config(self) -> Dict[str, Any]:
        return {'id': self.layer_id,
                'type': self.layer_type,
                'config': self.config,
                'visualChannels': self.visual_channels
                }


class GeoMapLayer(MapLayer):
    def __init__(self, data_id: str, label: str, color_column: str):
        self._id = MapLayer.generate_id()
        self._data_id = data_id
        self._label = label
        self._color_column = color_column

    @property
    def config(self) -> Dict[str, Any]:
        return {'dataId': self._data_id,
                'label': self._label,
                'color': [137, 218, 193],
                'highlightColor': HIGHLIGHT_COLOR,
                'columns': {'geojson': 'geometry'},
                'isVisible': True,
                'visConfig': {'opacity': 0.6,
                              'strokeOpacity': 0.8,
                              'thickness': 0.5,
                              'strokeColor': [179, 173, 158],
                              'colorRange': COLOR_RANGE,
                              'enableElevationZoomFactor': True,
                              'stroked': True,
                              'filled': True,
                              'enable3d': False,
                              'wireframe': False},
                'hidden': False}

    @property
    def visual_channels(self) -> Dict[str, Any]:
        return {'colorField': {'name': self._color_column,
                               'type': 'real'},
                'colorScale': 'quantile',
                'strokeColorField': None,
                'strokeColorScale': 'quantile',
                'sizeField': None,
                'sizeScale': 'linear',
                'heightField': None,
                'heightScale': 'linear',
                'radiusField': None,
                'radiusScale': 'linear'}

    @property
    def layer_type(self) -> str:
        return 'geojson'


class ArcLayer(MapLayer):
    def __init__(self, data_id: str, label: str, origin_lat: str, origin_lng: str,
                 destination_lat: str, destination_lng: str, size_column: str):
        self._id = MapLayer.generate_id()
        self._data_id = data_id
        self._label = label
        self._origin_lat = origin_lat
        self._origin_lng = origin_lng
        self._destination_lat = destination_lat
        self._destination_lng = destination_lng
        self._size_column = size_column

    @property
    def layer_type(self) -> str:
        return 'arc'

    @property
    def config(self) -> Dict[str, Any]:
        return {'dataId': self._data_id,
                'label': self._label,
                'color': [34, 63, 154],
                'highlightColor': HIGHLIGHT_COLOR,
                'columns': {'lat0': self._origin_lat,
                            'lng0': self._origin_lng,
                            'lat1': self._destination_lat,
                            'lng1': self._destination_lng},
                'isVisible': True,
                'visConfig': {'opacity': 0.8,
                              'thickness': 2,
                              'colorRange': COLOR_RANGE,
                              'sizeRange': [0, 10],
                              'targetColor': None},
                'hidden': False,
                'textLabel': [{'field': None,
                               'color': [255, 255, 255],
                               'size': 18,
                               'offset': [0, 0],
                               'anchor': 'start',
                               'alignment': 'center'}]}

    @property
    def visual_channels(self) -> Dict[str, Any]:
        return {'colorField': None,
                'colorScale': 'quantile',
                'sizeField': {'name': self._size_column, 'type': 'real'},
                'sizeScale': 'linear'}


class HeatMapLayer(MapLayer):
    def __init__(self, data_id: str, label: str, lat_column: str, lng_column: str, weight_column: str):
        self._id = MapLayer.generate_id()
        self._data_id = data_id
        self._label = label
        self._lat = lat_column
        self._lng = lng_column
        self._weight_column = weight_column

    @property
    def layer_type(self) -> str:
        return 'heatmap'

    @property
    def config(self) -> Dict[str, Any]:
        return {'dataId': self._data_id,
                'label': self._label,
                'color': [130, 154, 227],
                'highlightColor': HIGHLIGHT_COLOR,
                'columns': {'lat': self._lat,
                            'lng': self._lng},
                'isVisible': True,
                'visConfig': {'opacity': 0.8,
                              'colorRange': COLOR_RANGE,
                              'radius': DEFAULT_HEATMAP_RADIUS},
                'hidden': False,
                'textLabel': [{'field': None,
                               'color': [255, 255, 255],
                               'size': 18,
                               'offset': [0, 0],
                               'anchor': 'start',
                               'alignment': 'center'}]}

    @property
    def visual_channels(self) -> Dict[str, Any]:
        return {'weightField': {'name': self._weight_column, 'type': 'real'},
                'weightScale': 'linear'}



import pytest
import pandas as pd
import geopandas as gpd

from shapely.geometry import Point
from sttn.network import SpatioTemporalNetwork

edges = {'origin': [1, 2], 'destination': [3, 4]}
edges_pd = pd.DataFrame(data=edges)

nodes = {'id': [1, 2], 'geometry': [Point(1, 2), Point(2, 1)]}
nodes_gpd = gpd.GeoDataFrame(nodes, crs="EPSG:4326")
nodes_indexed = nodes_gpd.set_index('id')


def test_data_type_errors():
    # invalid nodes type
    with pytest.raises(TypeError):
        SpatioTemporalNetwork(nodes=edges_pd, edges=edges_pd)
    # invalid edges type
    with pytest.raises(TypeError):
        SpatioTemporalNetwork(nodes=nodes_gpd, edges=edges)

    # inconsistent origin/destination types
    origin_casted = edges_pd.astype({'origin': 'int32'})
    with pytest.raises(TypeError):
        SpatioTemporalNetwork(nodes=nodes_indexed, edges=origin_casted)

    # inconsistent index types
    od_casted = edges_pd.astype({'origin': 'int32', 'destination': 'int32'})
    with pytest.raises(TypeError):
        SpatioTemporalNetwork(nodes=nodes_indexed, edges=od_casted)


def test_column_name_errors():
    with pytest.raises(KeyError):
        SpatioTemporalNetwork(nodes=nodes_gpd, edges=edges_pd, origin='origin_2')
    with pytest.raises(KeyError):
        SpatioTemporalNetwork(nodes=nodes_gpd, edges=edges_pd, destination='destination_2')


def test_index_check_errors():
    # no node index
    with pytest.raises(ValueError):
        SpatioTemporalNetwork(nodes=nodes_gpd, edges=edges_pd)


def test_index_values():
    # destination ids are not in the node index
    with pytest.raises(KeyError):
        SpatioTemporalNetwork(nodes=nodes_indexed, edges=edges_pd)



import json
import requests
import os
import pathlib
import hashlib


class DataProvider:
    CHUNK_SIZE = 1 << 20

    def cache_file(self, url, local_filename=None):
        if not local_filename:
            local_filename = url.split('/')[-1]

        pathlib.Path(self.cache_dir()).mkdir(parents=True, exist_ok=True)
        file_path = os.path.join(self.cache_dir(), local_filename)

        if not os.path.exists(file_path):
            with requests.get(url, stream=True) as r:
                r.raise_for_status()
                with open(file_path, 'wb') as f:
                    for chunk in r.iter_content(chunk_size=self.CHUNK_SIZE):
                        f.write(chunk)
        return file_path

    def cache_dir(self):
        return os.path.join(pathlib.Path.home(), '.sttn', 'data', self.__class__.__name__)

    @staticmethod
    def hash_args(**kwargs) -> str:
        """MD5 hash of a dictionary."""
        md5 = hashlib.md5()
        encoded = json.dumps(kwargs, sort_keys=True).encode()
        md5.update(encoded)
        return md5.hexdigest()



from sttn import constants
import networkx as nx
import pandas as pd
import geopandas as gpd
import skmob


class SpatioTemporalNetwork:
    def __init__(self, nodes: gpd.GeoDataFrame, edges: pd.DataFrame, origin: str = constants.ORIGIN,
                 destination: str = constants.DESTINATION, node_id: str = constants.NODE_ID):
        if not isinstance(nodes, gpd.GeoDataFrame):
            raise TypeError('Incompatible nodes data type: {e}'.format(e=type(edges)))

        if not isinstance(edges, pd.DataFrame):
            raise TypeError('Incompatible edges data type: {e}'.format(e=type(edges)))

        if origin not in edges:
            raise KeyError('Origin column name: {orig} is not found in the list: {columns}'
                           .format(orig=origin, columns=list(edges.columns)))

        if destination not in edges:
            raise KeyError('Destination column name: {dest} is not found in the list: {columns}'
                           .format(dest=destination, columns=list(edges.columns)))

        if not nodes.index.name == node_id:
            raise ValueError('Nodes dataframe must be indexed on {id}'.format(id=node_id))

        if edges[origin].dtype != edges[destination].dtype:
            raise TypeError('Origin dtype {o} does not match destination dtype {d}'
                            .format(o=edges[origin].dtype, d=edges[destination].dtype))

        if edges[origin].dtype != nodes.index.dtype:
            raise TypeError('Origin dtype {o} does not match node index dtype {d}'
                            .format(o=edges[origin].dtype, d=nodes.index.dtype))

        SpatioTemporalNetwork._validate_ids(edges[origin], nodes.index)
        SpatioTemporalNetwork._validate_ids(edges[destination], nodes.index)

        self._nodes = nodes
        self._edges = edges
        self._origin = origin
        self._destination = destination
        self._node_id = node_id

    @staticmethod
    def _validate_ids(edge_ids: pd.Series, node_index: pd.Index):
        not_in_index = edge_ids[~edge_ids.isin(node_index)]
        if not_in_index.size > 0:
            samples = not_in_index.unique()[:5]
            raise KeyError('Edge ids {ids} are not in the node index'.format(ids=samples))

    @property
    def nodes(self) -> gpd.GeoDataFrame:
        return self._nodes

    @property
    def edges(self) -> pd.DataFrame:
        return self._edges

    def agg_parallel_edges(self, column_aggs: dict, key: str = None):
        grouping = [self._origin, self._destination]
        if key:
            grouping.append(key)
        new_edges = self._edges.groupby(by=grouping, as_index=False).agg(column_aggs)
        return SpatioTemporalNetwork(nodes=self._nodes, edges=new_edges, origin=self._origin,
                                     destination=self._destination, node_id=self._node_id)

    def to_multigraph(self):
        return nx.from_pandas_edgelist(self._edges, source=self._origin, target=self._destination,
                                       edge_attr=True, create_using=nx.MultiDiGraph)

    def to_flow_date_frame(self, flow: str) -> skmob.FlowDataFrame:
        return skmob.FlowDataFrame(self._edges, origin=self._origin, destination=self._destination, flow=flow,
                                   tile_id=self._node_id, tessellation=self._nodes.reset_index())

    def shape(self) -> (int, int):
        return self._nodes.shape[0], self._edges.shape[0]

    def group_nodes(self, node_label):
        nodes = self._nodes

        if isinstance(node_label, list):
            node_to_label = [(item, ind) for ind, sublist in enumerate(node_label) for item in sublist]
            community_df = pd.DataFrame(node_to_label, columns=[self._node_id, 'community'])
            community_df = community_df.set_index(self._node_id)
            nodes = self._nodes.join(community_df)
            node_label = 'community'

        dissolved = nodes.dissolve(by=node_label, as_index=False).rename(columns={node_label: self._node_id})
        dissolved = dissolved.set_index(self._node_id)

        node_mapping = nodes[[node_label]]
        mapped_from = self._edges.join(node_mapping, on=self._origin)\
            .drop(self._origin, axis=1)\
            .rename(columns={node_label: self._origin})
        mapped_to = mapped_from.join(node_mapping, on=self._destination)\
            .drop(self._destination, axis=1)\
            .rename(columns={node_label: self._destination})
        return SpatioTemporalNetwork(nodes=dissolved, edges=mapped_to, origin=self._origin,
                                     destination=self._destination, node_id=self._node_id)

    def agg_adjacent_edges(self, aggs: dict, outgoing: bool = True, include_cycles: bool = True) -> pd.DataFrame:
        grouping_column = self._origin if outgoing else self._destination
        edges = self._edges if include_cycles else self._edges[self._edges[self._origin] != self._edges[self._destination]]
        grouped = edges.groupby(grouping_column).agg(aggs)
        return grouped.rename(columns={grouping_column: self._node_id})

    def join_node_labels(self, extra_columns):
        new_nodes = self._nodes.join(extra_columns)
        return SpatioTemporalNetwork(nodes=new_nodes, edges=self._edges, origin=self._origin,
                                     destination=self._destination, node_id=self._node_id)

    def filter_nodes(self, condition: pd.Series):
        if self._nodes.shape[0] != condition.count():
            msg = 'Number of nodes {nodes} is different from the length of the condition array {condition}'.format(
                nodes=self._nodes.shape[0], condition=condition.count())
            raise ValueError(msg)

        ids_to_keep = self._nodes[condition].index
        filtered_edges = self._edges[self._edges[self._origin].isin(ids_to_keep) & self._edges[self._destination].isin(ids_to_keep)]
        return SpatioTemporalNetwork(nodes=self._nodes[condition], edges=filtered_edges, origin=self._origin,
                                     destination=self._destination, node_id=self._node_id)

    def filter_edges(self, condition: pd.Series):
        if self._edges.shape[0] != condition.count():
            msg = 'Number of edges {edges} is different from the length of the condition array {condition}'.format(
                edges=self._edges.shape[0], condition=condition.count())
            raise ValueError(msg)

        filtered_edges = self._edges[condition]
        return SpatioTemporalNetwork(nodes=self._nodes, edges=filtered_edges, origin=self._origin,
                                     destination=self._destination, node_id=self._node_id)



import pandas as pd

from sttn import network
from .data_provider import DataProvider
from . import census


class OriginDestinationEmploymentDataProvider(DataProvider):
    """
    Longitudinal Employer-Household Dynamics Origin-Destination Employment Statistics
    Data spec: https://lehd.ces.census.gov/data/lodes/LODES7/LODESTechDoc7.5.pdf
    """

    def build_network(self, state: str, year: int) -> network.SpatioTemporalNetwork:
        od_columns = ['w_geocode', 'h_geocode', 'S000', 'SA01', 'SA02', 'SA03', 'SE01', 'SE02', 'SE03', 'SI01', 'SI02',
                      'SI03']
        xwalk_columns = ['tabblk2010', 'trct', 'ctyname', 'zcta']
        od_data = pd.read_csv(self.od_fname, compression='gzip', usecols=od_columns)
        xwalk_data = pd.read_csv(self.xwalk_fname, compression='gzip', index_col='tabblk2010', usecols=xwalk_columns)
        # map census Block Codes to Census Tract codes
        home_joined = xwalk_data[['trct']].merge(od_data, left_index=True, right_on='h_geocode', how='inner').rename(
            columns={'trct': 'origin'})
        joined = xwalk_data[['trct']].merge(home_joined, left_index=True, right_on='w_geocode', how='inner').rename(
            columns={'trct': 'destination'})
        cleaned = joined.drop(['w_geocode', 'h_geocode'], axis=1)
        aggregated_edges = cleaned.groupby(['origin', 'destination']).sum().reset_index()

        rename_map = {'trct': 'id', 'ctyname': 'county', 'zcta': 'zip'}
        renamed = xwalk_data.reset_index()[['trct', 'ctyname', 'zcta']].rename(columns=rename_map)
        # 99999 is used for unknown zip codes
        tract_to_zip = renamed[renamed.zip != 99999].groupby('id').first()
        tract_to_zip['zip'] = tract_to_zip['zip'].astype(str)
        tract_geo_columns = ['GEOID', 'geometry']
        tract_shapes = census.get_tract_geo(state=state, year=year)
        # filter out water-only tracts:
        filtered_tracts = tract_shapes[tract_shapes.ALAND > 0]
        indexed_tracts = filtered_tracts[tract_geo_columns].set_index('GEOID')
        tracts_with_zip = indexed_tracts.merge(tract_to_zip, left_index=True, right_on='id', how='inner')

        # filter out edges for filtered nodes
        ids_to_keep = tracts_with_zip.index
        filtered_edges = aggregated_edges[aggregated_edges.origin.isin(ids_to_keep) & aggregated_edges.destination.isin(ids_to_keep)]
        return network.SpatioTemporalNetwork(nodes=tracts_with_zip, edges=filtered_edges)

    def get_data(self, state: str, year: int, part: str = 'main', job_type: int = 0) -> network.SpatioTemporalNetwork:
        self._cache(state=state, year=year, part=part, job_type=job_type)
        return self.build_network(state=state, year=year)

    def _cache(self, state: str, year: int, part: str = 'main', job_type: int = 0) -> None:
        od_fname = '{state}_od_{part}_JT0{job_type}_{year}.csv.gz'.format(
            state=state, part=part, job_type=job_type, year=year)
        xwalk_fname = '{state}_xwalk.csv.gz'.format(state=state)

        state_url = 'https://lehd.ces.census.gov/data/lodes/LODES7/' + state
        od_url = state_url + '/od/' + od_fname
        self.od_fname = self.cache_file(od_url)

        xwalk_url = state_url + '/' + xwalk_fname
        self.xwalk_fname = self.cache_file(xwalk_url)



import pandas as pd
from sttn.network import SpatioTemporalNetwork

from haversine import haversine_vector, Unit
from networkx.algorithms import community


def get_edges_with_centroids(network: SpatioTemporalNetwork) -> pd.DataFrame:
    centroid = network.nodes.centroid
    centroid_long = centroid.x
    centroid_long.name = 'long'
    centroid_lat = centroid.y
    centroid_lat.name = 'lat'
    centroids = pd.concat([centroid_long, centroid_lat], axis=1)
    centroid_from = network.edges.join(centroids, on=network._origin).rename(
        columns={'long': 'long_from', 'lat': 'lat_from'})
    centroid_all = centroid_from.join(centroids, on=network._destination).rename(
        columns={'long': 'long_to', 'lat': 'lat_to'})
    return centroid_all


def add_distance(network: SpatioTemporalNetwork) -> SpatioTemporalNetwork:
    """Add distance in km between area centroids."""
    centroid_all = get_edges_with_centroids(network)
    from_points = list(zip(centroid_all.lat_from, centroid_all.long_from))
    to_points = list(zip(centroid_all.lat_to, centroid_all.long_to))
    centroid_all['distance'] = haversine_vector(from_points, to_points, Unit.KILOMETERS)
    centroid_all.drop(['long_from', 'lat_from', 'long_to', 'lat_to'], axis=1, inplace=True)
    return SpatioTemporalNetwork(nodes=network.nodes, edges=centroid_all)


def detect_communities(network: SpatioTemporalNetwork, algo, **kwargs):
    if algo == 'fluid':
        comm_iter = community.asyn_fluidc(network.to_multigraph().to_undirected(), **kwargs)
        return list(comm_iter)
    if algo == 'clm':
        comm_iter = community.greedy_modularity_communities(network.to_multigraph().to_undirected(), **kwargs)
        return list(comm_iter)



import pytest
import pandas as pd
import geopandas as gpd

from pandas.testing import assert_frame_equal

from shapely.geometry import Point
from sttn.network import SpatioTemporalNetwork

edges = {'origin': [1, 1, 2, 1, 2], 'destination': [2, 2, 1, 2, 2], 'value': [1, 2, 4, 8, 16], 'key': [1, 2, 1, 1, 1]}
edges_pd = pd.DataFrame(data=edges)

nodes = {'id': [1, 2, 3], 'geometry': [Point(1, 2), Point(2, 1), Point(3, 1)]}
nodes_gpd = gpd.GeoDataFrame(nodes, crs="EPSG:4326").set_index('id')
stn = SpatioTemporalNetwork(nodes=nodes_gpd, edges=edges_pd)


def test_agg_parallel_edges_no_key():
    aggregated = stn.agg_parallel_edges(column_aggs={'value': 'sum'})
    assert aggregated.edges.shape[0] == 3
    assert set(aggregated.edges.columns) == {'origin', 'destination', 'value'}

    expected = pd.DataFrame(data={'origin': [1, 2, 2], 'destination': [2, 1, 2], 'value': [11, 4, 16]})
    assert_frame_equal(aggregated.edges, expected)
    assert_frame_equal(aggregated.nodes, nodes_gpd)


def test_agg_parallel_edges_with_key():
    aggregated = stn.agg_parallel_edges(column_aggs={'value': 'sum'}, key='key')
    assert aggregated.edges.shape[0] == 4
    assert set(aggregated.edges.columns) == {'origin', 'destination', 'value', 'key'}

    expected = pd.DataFrame(data={'origin': [1, 1, 2, 2], 'destination': [2, 2, 1, 2], 'key': [1, 2, 1, 1], 'value': [9, 2, 4, 16]})
    assert_frame_equal(aggregated.edges, expected)


def test_agg_adjacent_edges():
    # outgoing = True, include_cycles = True
    edges_aggregated = stn.agg_adjacent_edges(aggs={'value': 'sum'})
    expected_outgoing = pd.DataFrame(data={'origin': [1, 2], 'value': [11, 20]}).set_index('origin')
    assert_frame_equal(edges_aggregated, expected_outgoing)

    # outgoing = False, include_cycles = True
    incoming_aggregated = stn.agg_adjacent_edges(aggs={'value': 'sum'}, outgoing=False)
    expected_incoming = pd.DataFrame(data={'destination': [1, 2], 'value': [4, 27]}).set_index('destination')
    assert_frame_equal(incoming_aggregated, expected_incoming)

    # outgoing = True, include_cycles = False
    aggregated_no_cycle = stn.agg_adjacent_edges(aggs={'value': 'sum'}, include_cycles=False)
    expected_no_cycle = pd.DataFrame(data={'origin': [1, 2], 'value': [11, 4]}).set_index('origin')
    assert_frame_equal(aggregated_no_cycle, expected_no_cycle)



from abc import ABC, abstractmethod
from typing import Any, Dict, List, Tuple

import random
import string


class MapFilter(ABC):
    _id: str
    _data_id: List[str]

    @property
    def filter_id(self) -> str:
        return self._id

    @property
    def data_id(self) -> List[str]:
        return self._data_id

    @staticmethod
    def generate_id() -> str:
        return ''.join(random.choices(string.ascii_lowercase, k=5))

    @property
    @abstractmethod
    def filter_type(self) -> str:
        return NotImplemented

    @abstractmethod
    def extra_config(self) -> Dict[str, Any]:
        return NotImplemented

    def get_config(self) -> Dict[str, Any]:
        config = {'dataId': self._data_id,
                  'id': self.filter_id,
                  'type': self.filter_type,
                  }
        config.update(self.extra_config())

        return config


class TimeRangeFilter(MapFilter):
    def __init__(self, data_id: str, time_column: str, window: Tuple[int, int], y_axis_column: str):
        self._id = MapFilter.generate_id()
        self._data_id = [data_id]
        self._time_column = time_column
        self._time_window = [window[0], window[1]]
        self._y_axis_column = y_axis_column

    @property
    def filter_type(self) -> str:
        return 'timeRange'

    def extra_config(self) -> Dict[str, Any]:
        return {'name': [self._time_column], 'value': self._time_window, 'enlarged': True,
                'plotType': 'lineChart',
                'animationWindow': 'free',
                'yAxis': {'name': self._y_axis_column, 'type': 'integer'},
                'speed': 1}



from __future__ import annotations
from typing import Any, Dict, Tuple

from sttn.plot.keplergl.layer import ArcLayer, GeoMapLayer, HeatMapLayer
from sttn.plot.keplergl.filter import TimeRangeFilter


class MapConfig:
    def __init__(self, data_id: str):
        self.layers = []
        self.filters = []
        self._data_id = data_id

    def with_data_id(self, data_id) -> MapConfig:
        self._data_id = data_id
        return self

    def add_geo_layer(self, label: str, color_column: str) -> MapConfig:
        layer = GeoMapLayer(data_id=self._data_id, label=label, color_column=color_column)
        self.layers.append(layer)
        return self

    def add_arc_layer(self, label: str, origin_lat: str, origin_lng: str,
                      destination_lat: str, destination_lng: str, size_column: str) -> MapConfig:
        layer = ArcLayer(data_id=self._data_id, label=label, origin_lat=origin_lat, origin_lng=origin_lng,
                         destination_lat=destination_lat, destination_lng=destination_lng, size_column=size_column)
        self.layers.append(layer)
        return self

    def add_heatmap_layer(self, label: str, lat_column: str, lng_column: str, weight_column: str):
        layer = HeatMapLayer(data_id=self._data_id, label=label, lat_column=lat_column, lng_column=lng_column,
                             weight_column=weight_column)
        self.layers.append(layer)
        return self

    def add_time_range_filter(self, time_column: str, window: Tuple[int, int], y_axis_column: str) -> MapConfig:
        range_filter = TimeRangeFilter(data_id=self._data_id, time_column=time_column, window=window,
                                       y_axis_column=y_axis_column)
        self.filters.append(range_filter)
        return self

    def to_dict(self) -> Dict[str, Any]:
        layer_configs = [layer.get_config() for layer in self.layers]
        filter_configs = [map_filter.get_config() for map_filter in self.filters]
        return {'version': 'v1', 'config': {'visState': {'filters': filter_configs,
                                                         'layers': layer_configs}}}
